from datetime import datetime
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.db import models
from wagtail.admin.panels import FieldPanel
from wagtail.fields import RichTextField
from wagtail.models import Page
from wagtail.snippets.models import register_snippet


#
# MoNkEyPaTcHiNg TiMe
#

# bypasses form validation to allow blank slugs on submit
from wagtail.admin.forms.models import WagtailAdminModelForm


def clean(self):
    cleaned_data = super(WagtailAdminModelForm, self).clean()
    if isinstance(self.instance, Page):
        if 'slug' in self.errors and not self.data['slug']:
            cleaned_data['slug'] = ''
            del self.errors['slug']
    return cleaned_data


WagtailAdminModelForm.clean = clean


# allows hitting the db for authorized posts only
# instead of filtering in python
from wagtail.models import PageViewRestriction, PageQuerySet


def authorized_q(self, request):
    PASSWORD = PageViewRestriction.PASSWORD
    LOGIN    = PageViewRestriction.LOGIN
    GROUPS   = PageViewRestriction.GROUPS
    KEY      = PageViewRestriction.passed_view_restrictions_session_key

    restrictions = PageViewRestriction.objects.all()
    passed = request.session.get(KEY, [])
    if passed:
        restrictions = restrictions.exclude(id__in=passed,
                                            restriction_type=PASSWORD)
    if request.user.is_authenticated:
        restrictions = restrictions.exclude(restriction_type=LOGIN)
    if request.user.is_superuser:
        restrictions = restrictions.exclude(restriction_type=GROUPS)
    else:
        membership = request.user.groups.all()
        if membership:
            restrictions = restrictions.exclude(groups__in=membership,
                                                restriction_type=GROUPS)
    q = models.Q()
    for restriction in restrictions:
        q &= ~self.descendant_of_q(restriction.page, inclusive=True)
    return q


def authorized(self, request):
    self.request = request
    if request is None:
        return self
    else:
        return self.filter(self.authorized_q(request))


PageQuerySet.authorized_q = authorized_q
PageQuerySet.authorized = authorized
# monkeypatch party over


class PostMixin:
    subpage_types = []

    def index_snippet(self):
        return f'blog/_{self.content_type.model[:-4]}_index.html'

    def icon(self):
        return f'icons/{self.content_type.model[:-4]}.svg'


class NotePage(PostMixin, Page):
    body = RichTextField(max_length=500)
    
    content_panels = [
        FieldPanel('body'),
    ]
    promote_panels = Page.promote_panels[1:]

    parent_page_types = ['NoteIndexPage']

    def save(self, *args, **kwargs):
        if not self.title:
            date = datetime.now()
            self.title = datetime.strftime(date, '%Y%m%d-%H%M%S-%f')
        super().save(*args, **kwargs)


class ArticlePage(PostMixin, Page):
    body = RichTextField()

    content_panels = Page.content_panels + [
        FieldPanel('body'),
    ]

    parent_page_types = ['ArticleIndexPage']


class PhotoPage(PostMixin, Page):
    photo = models.ForeignKey(
        'wagtailimages.Image',
        on_delete=models.PROTECT,
    )

    content_panels = Page.content_panels + [
        FieldPanel('photo'),
    ]

    parent_page_types = ['PhotoIndexPage']


class PostIndexPage(Page):
    show_in_menus_default = True

    blurb = models.TextField(blank=True, null=True)
    posts_per_page = models.IntegerField(default=15)

    parent_page_types = ['blog.RootIndexPage']

    content_panels = Page.content_panels + [
        FieldPanel('blurb'),
        FieldPanel('posts_per_page'),
    ]

    def get_template(self, *args, **kwargs):
        return 'blog/post_index_page.html'

    def get_context(self, request):
        context = super().get_context(request)
        posts = self.get_descendants() \
                    .authorized(request) \
                    .live() \
                    .filter(show_in_menus=False) \
                    .order_by('-first_published_at')
        for ptype in ['note', 'article', 'photo']:
            count = posts.filter(content_type__model=f'{ptype}page').count()
            context[f'{ptype}_count'] = count
        page = request.GET.get('p')
        paginator = Paginator(posts, self.posts_per_page)
        try:
            posts = paginator.page(page)
        except PageNotAnInteger:
            posts = paginator.page(1)
        except EmptyPage:
            posts = paginator.page(paginator.num_pages)
        context['posts'] = posts
        return context


class RootIndexPage(PostIndexPage):
    subpage_types = [
        'NoteIndexPage',
        'ArticleIndexPage',
        'PhotoIndexPage',
    ]

    parent_page_types = [Page]


class NoteIndexPage(PostIndexPage):
    parent_page_types = ['RootIndexPage']
    subpage_types = ['NotePage']


class ArticleIndexPage(PostIndexPage):
    parent_page_types = ['RootIndexPage']
    subpage_types = ['ArticlePage']


class PhotoIndexPage(PostIndexPage):
    parent_page_types = ['RootIndexPage']
    subpage_types = ['PhotoPage']


@register_snippet
class Tagline(models.Model):
    text = models.CharField(max_length=255)

    panels = [
        FieldPanel("text"),
    ]

    def __str__(self):
        return self.text
